// Copyright (C) 2011 by the BEM++ Authors
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "assembly/assembly_options.hpp"
#include "assembly/boundary_operator.hpp"
#include "assembly/context.hpp"
#include "assembly/evaluation_options.hpp"
#include "assembly/grid_function.hpp"
#include "assembly/l2_norm.hpp"
#include "assembly/numerical_quadrature_strategy.hpp"
#include "assembly/surface_normal_independent_function.hpp"

#include "assembly/identity_operator.hpp"
#include "assembly/helmholtz_3d_single_layer_boundary_operator.hpp"
#include "assembly/helmholtz_3d_double_layer_boundary_operator.hpp"
#include "assembly/helmholtz_3d_single_layer_potential_operator.hpp"
#include "assembly/helmholtz_3d_double_layer_potential_operator.hpp"

#include "common/boost_make_shared_fwd.hpp"

#include "grid/grid.hpp"
#include "grid/grid_factory.hpp"

#include "linalg/default_iterative_solver.hpp"

#include "space/piecewise_linear_continuous_scalar_space.hpp"
#include "space/piecewise_constant_scalar_space.hpp"

#include <iostream>
#include <fstream>
#include <ctime>

typedef double BFT; // basis function type
typedef std::complex<double> RT; // result type (type used to represent discrete operators)
typedef double CT; // coordinate type

// The Neumann boundary condition
class NeumannData
{
public:
	// Type of the function's values (e.g. float or std::complex<double>)
    typedef RT ValueType;
    // Type of coordinates (must be the "real part" of ValueType)
    typedef CT CoordinateType;

    // Number of components of the function's argument
    int argumentDimension() const { return 3; }
    // Number of components of the function's value
    int resultDimension() const { return 1; }

    // Constructor
	NeumannData(const RT& k) : k(k)
	{
	}
	
	// Evaluate (constant function)
	inline void evaluate(const arma::Col<CoordinateType>& point,
                         arma::Col<ValueType>& result) const {
    
		// Note: currently result is independent of wave number
        result(0) = RT(1.0, 0.0);
    }
private:
	// The wave number
	RT k;
};

// The DirichletData to evaluate precision
class DirichletData
{
public:
    // Type of the function's values (e.g. float or std::complex<double>)
    typedef RT ValueType;
    // Type of coordinates (must be the "real part" of ValueType)
    typedef CT CoordinateType;

    // Number of components of the function's argument
    int argumentDimension() const { return 3; }
    // Number of components of the function's value
    int resultDimension() const { return 1; }

    // 
	DirichletData(const RT& k) : k(k)
	{
	}
    
    // Evaluate the function at the point "point" and store result in
    // the array "result"
    inline void evaluate(const arma::Col<CoordinateType>& point,
                         arma::Col<ValueType>& result) const {
        CoordinateType x = point(0), y = point(1), z = point(2);
        CoordinateType r = sqrt(point(0) * point(0) +
				point(1) * point(1) +
				point(2) * point(2));
        
		RT _i = RT(0.0, 1.0);				// Imaginary unit
        result(0) = -r / (1.0 - _i*k*r);	// Analytic pressure result
    }
private:
	RT k;	// The wave number
};

/**
 * The main function
 */
int main(int argc, char* argv[])
{
    // Import symbols from namespace Bempp to the global namespace
    using namespace Bempp;

	// Check number of input arguments
	if (argc < 3)
	{
		std::cout << "Usage: tutorial_helmholtz <mesh file name> <k>" << std::endl;
		return -1;
	}
	
	// Set the wave number
	RT k(atof(argv[2]), 0);
	
	// Load mesh
    const char* meshFile = argv[1];			// Mesh file name
    GridParameters params;
    params.topology = GridParameters::TRIANGULAR;
    shared_ptr<Grid> grid = GridFactory::importGmshGrid(params, meshFile);

	// Clock starts after mesh building
	std::cout << "Clock starts now!" << std::endl;
	std::clock_t begin = std::clock();
	
    // Initialize the spaces

    PiecewiseLinearContinuousScalarSpace<BFT> pwiseLinears(grid);
    PiecewiseConstantScalarSpace<BFT> pwiseConstants(grid);

    // Define the quadrature strategy

    AccuracyOptions accuracyOptions;
    // Increase by 2 the order of quadrature rule used to approximate
    // integrals of regular functions on pairs on elements
    accuracyOptions.doubleRegular.setRelativeQuadratureOrder(0);
    // Increase by 2 the order of quadrature rule used to approximate
    // integrals of regular functions on single elements
    accuracyOptions.singleRegular.setRelativeQuadratureOrder(0);
    NumericalQuadratureStrategy<BFT, RT> quadStrategy(accuracyOptions);

    // Specify the assembly method. We do not want to use ACA
    AssemblyOptions assemblyOptions;
	
	// NOTE: ACA assembly options are not used herein
    // AcaOptions acaOptions; // Default parameters for ACA
    // assemblyOptions.switchToAcaMode(acaOptions);

    // Create the assembly context
    Context<BFT, RT> context(make_shared_from_ref(quadStrategy), assemblyOptions);

    // Construct elementary operators
	BoundaryOperator<BFT, RT> slpOp =
		helmholtz3dSingleLayerBoundaryOperator<BFT>(
			make_shared_from_ref(context),
			make_shared_from_ref(pwiseLinears),
			make_shared_from_ref(pwiseLinears),		// pwiseLinears
			make_shared_from_ref(pwiseLinears),
			k);									
	
	BoundaryOperator<BFT, RT> dlpOp =
        helmholtz3dDoubleLayerBoundaryOperator<BFT>(
                make_shared_from_ref(context),
                make_shared_from_ref(pwiseLinears),	// pwiseLinears
                make_shared_from_ref(pwiseLinears),	// pwiseLinears
                make_shared_from_ref(pwiseLinears),
				k);
    
	BoundaryOperator<BFT, RT> idOp =
        identityOperator<BFT, RT>(
                make_shared_from_ref(context),
                make_shared_from_ref(pwiseLinears),	// pwiseLinears
                make_shared_from_ref(pwiseLinears),	// pwiseLinears
                make_shared_from_ref(pwiseLinears));

    // Form the right-hand side sum

    BoundaryOperator<BFT, RT> rhsOp = -0.5 * idOp + dlpOp;

    // Construct the grid function representing the (input) Neumann data

    GridFunction<BFT, RT> neumannData(
                make_shared_from_ref(context),
                make_shared_from_ref(pwiseLinears),	//pwiseLinears
                make_shared_from_ref(pwiseLinears),	//pwiseLinears
                surfaceNormalIndependentFunction(NeumannData(k)));

    // Construct the right-hand-side grid function
    GridFunction<BFT, RT> rhs = slpOp * neumannData;

    // Initialize the solver
    DefaultIterativeSolver<BFT, RT> solver(rhsOp);
    solver.initializeSolver(defaultGmresParameterList(1e-6));

    // Solve the equation
    Solution<BFT, RT> solution = solver.solve(rhs);
    std::cout << solution.solverMessage() << std::endl;

	std::clock_t end = clock();
	double elapsed_secs = double(end - begin) / double(CLOCKS_PER_SEC);
	std::cout << "Total time: " << elapsed_secs << " seconds." << std::endl;
	
    // Extract the solution in the form of a grid function
    // and export it in VTK format

	
    const GridFunction<BFT, RT>& solFun = solution.gridFunction();
	
    CT absoluteError, relativeError;
    estimateL2Error(
                solFun, surfaceNormalIndependentFunction(DirichletData(k)),
                quadStrategy, absoluteError, relativeError);
	
	// Display the error
    std::cout << std::endl << "Relative L^2 error: " << relativeError << std::endl;

	// Return success
	return 0;
}
