#include "core/weighted_residual.hpp"
#include "library/helmholtz_kernel.hpp"
#include "library/helmholtz_singular_integrals.hpp"
#include "library/lib_element.hpp"
#include "interface/read_off_mesh.hpp"

#include <chrono>
#include <iostream>
#include <cstdlib>
#include <Eigen/IterativeLinearSolvers>

typedef std::chrono::high_resolution_clock Clock;
typedef std::chrono::milliseconds milliseconds;

typedef Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> dMatrix;
typedef Eigen::Matrix<std::complex<double>, Eigen::Dynamic, Eigen::Dynamic> cMatrix;

int main(int argc, char *argv[])
{
	char const *fname = argv[1];
	double k = atof(argv[2]);

	auto mesh = read_off_mesh(fname, tria_1_tag());
	auto const &w = constant_view(mesh);

	auto N = w.get_num_dofs();
	cMatrix L(N, N);
	cMatrix M(N, N);
	L.setZero();
	M.setZero();

	auto LMop = create_integral_operator(
		helmholtz_3d_SLP_kernel<double>(k),
		helmholtz_3d_DLP_kernel<double>(k)
	);
	auto Iop = identity_integral_operator();

	Clock::time_point t0, t1;
	milliseconds ms;

	// compute matrix L
	t0 = Clock::now();
	(L, M) << dirac(w) * LMop[w];
	M << dirac(w) * (-.5*Iop)[w];
	t1 = Clock::now();
	ms = std::chrono::duration_cast<milliseconds>(t1 - t0);
	std::cout << ms.count() << " ms" << std::endl;

	// define BC
	cMatrix v(N, 1);
	v.setConstant(1.0);

	// solve system
	std::cout << "Starting iterative solution" << std::endl; 
	t0 = Clock::now();
	
	Eigen::BiCGSTAB<cMatrix> solver(M);
	solver.setTolerance(1e-6);
	cMatrix p = solver.solve(L*v);

	t1 = Clock::now();
	ms = std::chrono::duration_cast<milliseconds>(t1 - t0);
	std::cout << ms.count() << " ms" << std::endl;

	double r = 1.;
	std::complex<double> anal = -r/(1.+std::complex<double>(0.,1.)*k*r);
	dMatrix error(N,1);
	for (int i = 0; i < N; ++i)
		error(i) = std::abs(p(i)/anal-1.0);

	std::cout << "Mean error: " << error.sum()/N << std::endl;
	
	return 0;
}

